/*************************************************************

PROJECT   :  ULTRASONIC HAPTIC VISION

MEMBERS   :	 ROHAN SHARMA, JEFF BUENTE
CODE TYPE :  TRANSMITTER - MOTOR CONTROL, HAPTIC FEEDBACK AND RF-TX
CLASS	  :	 ECE 4760

**************************************************************/


//************************************************************
// Headers Files and other includes
//************************************************************

#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>
#include ".\txrx.c"

//************************************************************
// Defines
//************************************************************

//Timeout values for different segments of the code
#define tinit	3000	// Initialization time for sensor
#define tm 		10		// Time	for setting up a 1 mS counter using the Timer0 100 microSec counter
#define tdemo	300		// Time for executing each stage of the demo sequence

#define tmotor 	25		// Time for controlling the stepper motor
#define tsen	50		// Time for reading the range provided by the sensor
#define tvib	50		// Time for controlling the vibrator motors
#define trf		150		// Time for transmitting RF

//**********************************

// Motor Port defines
#define DDRM	DDRC
#define PORTM	PORTC
#define PINM	PINC

// Ultrasonic Sensor Port defines
#define DDRU	DDRA
#define PORTU	PORTA
#define PINU	PINA
#define p_input	0x01

// Vibrator Motor Port defines
#define DDRV	DDRB
#define PORTV	PORTB
#define PINV	PINB

// RF / Led Port defines
#define DDRL	DDRD
#define PORTL	PORTD
#define PINL	PIND
#define cpu_l	0x04	// Led output for CPU board
#define cal_i	0x10    // Input for Calibration
#define cal_o	0x20    // Output for Calibration

//**********************************

// Demo Defines
#define demoC		0x01
#define demoN		10

// Motor Defines
#define step1	0x0A	// Step 1 representation of stepper motor
#define step2	0x09	// Step 2 representation of stepper motor
#define step3	0x05	// Step 3 representation of stepper motor
#define step4	0x06	// Step 4 representation of stepper motor
#define inh1	0x10	// Enable for Bridge 1 on L298, always set to high
#define inh2	0x20	// Enable for Bridge 2 on L298, always set to high
#define inh		0x30	// Both enables combined together, always set to high
#define numSteps 	36	// Number of steps to take in 1 rotation

// Sensor Defines
#define rangeChoice	1	// 0-Analog, 1-PWM

// RF defines
#define rftx_data_length	6			// Maximum length of the RF payload
#define baudrate			4800		// Bauadrate for sending control signals to RF transmitted chip
#define f_cpu 				16000000UL	// Crystal frequency of the MCU
#define tx_id		 		0x0A		// Transmission ID of the RF packet
#define ccw_id				0x21		// ID for indicating a counter clockwise range packet
#define cw_id				0x12		// ID for indicating a clockwise range packet

// Vibrator motor defines
#define vib1	0			// Range index for vibrator motor 1	angle = -135 degrees
#define vib2	3			// Range index for vibrator motor 2 angle = -90 degrees
#define vib3	6			// Range index for vibrator motor 3 angle = -45 degrees
#define vib4	9			// Range index for vibrator motor 4 angle = 0 degree
#define vib5	12			// Range index for vibrator motor 5 angle = 45 degrees
#define vib6	15			// Range index for vibrator motor 6 angle = 90 degrees
#define vib7	18			// Range index for vibrator motor 7 angle = 135 degrees
#define p5ft		9       // 0.5 feet, used as a reference for calculating other ranges

//************************************************************
// Variable Declarations
//************************************************************

// Timer Counters for different segments/ modules of the program
volatile unsigned int time_init, time_m, time_motor, time_sen, time_rf, time_vib, time_demo; 

// Demo Declarations
unsigned char demoControl;			// Controls whether the program is in demo mode. Program in demo mode during startup for 3 secs
unsigned char demoNum;				// Specifies the number of times demo function needs to be executed 

// Calibration declaration
unsigned char calibrate;			// Calibration Check variable. 2-During initialization, 1-Calibration, 0-Calibrated 

// Motor declarations
unsigned char ccw;					// Rotation Control = 0-CounterClockwise, 1-Clockwise
unsigned char rot;  				// Step Counter. Specifies the step, and thus the angle at which motor is

// RF declarations
char rftx_data[rftx_data_length];	// Data to send over RF
char rf_range[3];					// Last 3 range values that need to be transmitted over RF
unsigned char packet_count; 		// Number of packets sent
unsigned char rf_id;				// Direction dependent ID for the RF packet 
unsigned long baudval;				// baud rate for RF signal transmission from UART0

// Sensor declarations
int  range;							// Range to be used for calculation
char rangeType;						// 0-Analog, 1-PWM
int  a_range; 						// Analog Distance Input from range sensor
int  p_range; 						// PWM Distnace Input from range sensor
char p_valid;						// Check if PWM range is valid or not
int  sen_range[19];					// Storing the range at all the angles
unsigned int senIndex;				// Index for storing the range at a particular angle

// Vibrator Control declarations
int vib_pwm[7];						// PWM width for each of the vibrator motors. This sets the intensity of vibration of motor.
int vib_index[7];					// Contains the angle index for each of the vibrators
int vib_range[7];					// Contains the range values for each of the vibrators
unsigned char vibLed;				// Contains the state of each of the vibrator motors (on/off). Used to control LEDs at receiver
int range1, range2, range3, range4; // Threshold ranges for setting the vibration intensity

//************************************************************
// Function Declarations
//************************************************************

void initialize(void); 		// Initialization for the entire MCU
void demo(void);			// A demo procedure of the vibrators
void rangeCalc(void);  		// Find range based on type of input
void motorControl(void);  	// Executes at time specified by t1
void rfTransmit(void);		// Create RF packet and transmit RF data
void vibControl(void);		// Controls the vibrator motor based on particular angle
void filterRange(void);		// Filter the range appropriately so that range values at angles are represented well

//************************************************************
// Interrupt Service Routines
//************************************************************

//Timer 0 compare A ISR
ISR (TIMER0_COMPA_vect)
{
  time_m++;

  //Decrement the time if not already zero
  if(time_m == tm)
  {
  	  time_m = 0;
	    
	  if (time_demo>0)	--time_demo;
	  if (time_init>0)	--time_init;
	  if (time_motor>0)	--time_motor;
	  if (time_rf>0)	--time_rf;
	  if (time_sen>0)	--time_sen;
	  if (time_vib>0)	--time_vib;	  

	  // Switch on all the motors at the start of the cycle
	  PORTV = 0xFF;	  
  }
  
  // Turn off the PWM for the vibrator motors at appropriate times  
  // If 0, then motor is turned off almost instantaneously
  if(time_m == vib_pwm[0] )		PORTV &= ~(0x01);
  if(time_m == vib_pwm[1] )		PORTV &= ~(0x02);
  if(time_m == vib_pwm[2] )		PORTV &= ~(0x04);
  if(time_m == vib_pwm[3] )		PORTV &= ~(0x08);
  if(time_m == vib_pwm[4] )		PORTV &= ~(0x10);
  if(time_m == vib_pwm[5] )		PORTV &= ~(0x20);
  if(time_m == vib_pwm[6] )		PORTV &= ~(0x40);

  // Calibration input, check when this input goes high and calibrate the motor position
  if((~PINL) & cal_i)
  {  	
	if(calibrate == 1)
	{
		// Set initial values of the motor
		ccw = 1;
		rot = 36;
		senIndex = 18; 

		// Turn motor to next step so that motor does not make contact with the wire in every rotation
		switch(PORTM & 0x0F)
		{
			case step1: PORTM = (inh | step4); break;
			case step2: PORTM = (inh | step1); break;
			case step3: PORTM = (inh | step2); break;
			case step4: PORTM = (inh | step3); break;
			default: PORTM = (inh | step1);
		}

		// Calibration complete indication
		calibrate = 0; 				
	}		
  }

  // Check if the PWM input from Ultrasonic sensor is high and start counting 
  if(PINU & p_input)
  {
  	if(p_valid == 1)
	{
		p_valid = 0;
		p_range = 0;
	}

	p_range++;	
  }
  else
  {
  	p_valid = 1;
	range = p_range;							
  }

}

//************************************************************
// main() function
//************************************************************

int main(void)
{

  // Initialize entire MCU
  initialize();
  
  // Main task scheduler loop
  while(1)
  {
    if((time_demo == 0) && (demoControl == 1))
	{
		// Executing the demo code
		demo();	
		
		demoNum--;

		if(demoNum == 0)
		{
			demoControl = 0;
			time_demo = 0;
		}
		else
		{			
			time_demo = tdemo;
		}		
	}
	
	// If demo executed, execute all other parts of the program
	if((time_init==0) && (demoControl==0))	
	{
		if(calibrate == 2)
		{
			calibrate  = 1;
		}

		if (time_motor==0)
	    {
	    	time_motor = tmotor;
			motorControl();			
	    }

		if(calibrate == 0)
		{
			if (time_sen == 0)
			{
				time_sen = tsen;
				rangeCalc();
			}
		
			if (time_vib == 0)
			{
				time_vib = tvib;
				vibControl();
			}

			if(time_rf == 0)
			{
				time_rf = trf;
				rfTransmit();			
				PORTL ^= (cpu_l);	
			}
		}
	}		
  }
}


//************************************************************
// Local Functions
//************************************************************

// demo() - Initial Demo Function
void demo(void)
{
	// Demo sequence steps. Vibrator motors start vibrating from one end, 
	// vibration progresses to the other end and returns to the first motor
	switch(demoNum)
	{
		case 10: vib_pwm[1] = 8; break;
		case 9:  vib_pwm[1] = 0; vib_pwm[2] = 8; break;
		case 8:  vib_pwm[2] = 0; vib_pwm[3] = 8; break;
		case 7:  vib_pwm[3] = 0; vib_pwm[4] = 8; break;
		case 6:  vib_pwm[4] = 0; vib_pwm[5] = 8; break;
		case 5:  vib_pwm[4] = 8; vib_pwm[5] = 0; break;
		case 4:  vib_pwm[3] = 8; vib_pwm[4] = 0; break;
		case 3:  vib_pwm[2] = 8; vib_pwm[3] = 0; break;
		case 2:  vib_pwm[1] = 8; vib_pwm[2] = 0; break;
		case 1:  vib_pwm[1] = 0; break;
	}
}	

//**************************************************

// filterRange() - Filter the range value at particular angles and the surrounding range values to represent the best 
// approximate vibrator range
void filterRange(void)
{

	int i;

	vib_range[0] = sen_range[vib_index[0]];
	vib_range[6] = sen_range[vib_index[6]];
		
	for(i=1; i<6; i++)
	{	
		vib_range[i] = ((2 * sen_range[vib_index[i]]) + sen_range[vib_index[i]-1] + sen_range[vib_index[i]+1]);
	}

	vib_range[3] = sen_range[vib_index[3]] << 2;

}

//**************************************************

// rangeCalc() - Calculate the range from the sensor depending on type of range
void rangeCalc(void)
{
	
	if(rangeType == 0)
	{
		a_range = ((int)ADCH);	
	  	ADCSRA |= (1<<ADSC); 

		range = a_range;
	}
	else if((rangeType == 1) && (p_valid == 1))
	{		
		range = p_range;						
	}
	
	
	rf_range[0] = rf_range[1]; 	
	rf_range[1] = rf_range[2];
	rf_range[2] = range;

	sen_range[senIndex] = range;

	if(ccw == 0)
	{
		senIndex++;
	}
	else if(ccw == 1)
	{
		senIndex--;
	}
}

//**************************************************

// vibControl() - scheduled by tvib. Used to control the operation of vibrator motors at particular angles
void vibControl(void)
{
	int i;

	vibLed = 0x00;
	
	for(i = 0; i<7; i++)
	{

		filterRange();			

		if(vib_range[i] == 0)
		{
			vib_pwm[i] = 0;
		}
		else if(vib_range[i] <= range1)
		{
			vib_pwm[i] = 9;
		}
		else if((vib_range[i] > range1) && (vib_range[i] <= range2))
		{
			vib_pwm[i] = 7;			
		}
		else if((vib_range[i] > range2) && (vib_range[i] <= range3))
		{
			vib_pwm[i] = 5;			
		}
		else if((vib_range[i] > range3) && (vib_range[i] <= range4))
		{
			vib_pwm[i] = 4;			
		}
		else if(vib_range[i] > range4)
		{
			vib_pwm[i] = 0;			
		}

		if(vib_pwm[i] <= 4)
		{
			vibLed |= (1 << i);
		}

	}

	vibLed = vibLed >> 1;
}


//**************************************************

// motorControl() - scheduled by tmotor, Generates control signals for stepper motor and controls rotation
// and number of steps
void motorControl(void)
{	
	if(ccw == 0)
	{
		// Counter clockwise step sequence
		switch(PORTM & 0x0F)
		{
			case step1: PORTM = (inh | step2); break;
			case step2: PORTM = (inh | step3); break;
			case step3: PORTM = (inh | step4); break;
			case step4: PORTM = (inh | step1); break;
			default: PORTM = (inh | step1);
		}
	}
	else if (ccw == 1)
	{
		// Clockwise step sequence
		switch(PORTM & 0x0F)
		{
			case step1: PORTM = (inh | step4); break;
			case step2: PORTM = (inh | step1); break;
			case step3: PORTM = (inh | step2); break;
			case step4: PORTM = (inh | step3); break;
			default: PORTM = (inh | step1);
		}
	}
	
	if(calibrate == 0)
	{
		if(ccw == 0)
		{ 
			rot++;
			if(rot == numSteps) 
			{
				ccw = 1;				
			}
		}
		else if(ccw == 1)
		{
			rot--;
			if(rot == 0) 
			{
				ccw = 0;					
			}
	    }
	}
}

//**************************************************

// Form and transmit RF data packet
void rfTransmit (void)
{
	// Increment packet count
	packet_count++;

	// Set the RF id according to rotation
	if(ccw == 0)
	{
		rf_id = ccw_id;
	}
	else if (ccw == 1)
	{
		rf_id = cw_id;
	}
	
	// Construct data packet
	rftx_data[0] = vibLed;					
	rftx_data[1] = rf_id;
	rftx_data[2] = rot;
	rftx_data[3] = rf_range[0];
	rftx_data[4] = rf_range[1];
	rftx_data[5] = rf_range[2];

	// Transmit data packet
	tx_me(rftx_data, rftx_data_length, tx_id);
}


//************************************************************
// MCU Initialization Functions
//************************************************************

// Primary Initialization Function
void initialize(void)
{
  // Set the type of the current range input from the sensor
  rangeType = rangeChoice;

  //ADC initialization  
  //Initialize the ADC only if the choice is set to 0
  if(rangeType == 0)	
  {
	  //init the A to D converter, channel zero, left adj, EXTERNAL Aref!
	  ADMUX = (1<<ADLAR) ;  
	  //enable ADC and set prescaler to 1/128*16MHz=125,000, and clear interupt enable, and start a conversion
	  ADCSRA = ((1<<ADEN) | (1<<ADSC)) + 7; 
  }	

//******************************************

  //Timer Initialization
  //Set up timer 0 for 1 mSec timebase
  TIMSK0= (1<<OCIE0A);							// turn on timer 0 cmp match ISR
  OCR0A = 24;  									// set the compare to 25 time ticks
  TCCR0B= 3; 									// set prescalar to divide by 64 //0b00000011;
  TCCR0A= (1<<WGM01) ;							// turn on clear-on-match
  
//******************************************

  //RF Initialization
  baudval = (f_cpu / (16UL * baudrate)) - 1; 
  txrx_init(1,0,(int)baudval, 0);				// Transmission initialization. Enable Transmit, Disable Receive, Set Baud Value

//******************************************

  //Port Initialization
  
  // Motor
  DDRM = 0xFF;					// Set all the stepper motor connections to outputs
  PORTM = (inh | step1);		// Set the initial step for the stepper motor

  // Ultrasonic Sensor
  DDRU &= ~(p_input);			// Set the sensor connection on port A as input

  // Vibrating Motor
  DDRV 	= 0xFF;					// Set all vibrator motor connections to outputs
  PORTV = 0x00;					// Turn off all vibrator motors initially

  // Leds
  DDRL  |= (cpu_l | cal_o);		// Set the CPU led and calibration output connections as outputs
  DDRL  &= ~(cal_i);			// Set the calibration input as input
  PORTL |= (cpu_l | cal_i);		// Turn off Led (active low) and pull up cal_i input
  PORTL &= ~(cal_o);			// Set cal_o output to 0

//******************************************

  // Variable Initialization

  // Initial Calibration
  calibrate = 2;
  
  // Timers
  time_init = tinit;
  time_m = 0;
  time_sen = 0;
  time_motor = 0;
  time_rf = 0;
  time_vib = 0;

  // Motors
  ccw = 0;
  rot = 0;
  
  // RF
  rf_id = ccw_id;
  packet_count = 0;

  // Ultrasonic Sensor
  senIndex = 0;
  range = 0;
  a_range = 0;
  p_range = 0;
  p_valid = 0;

  // Vibrator Motors
  vib_index[0] = vib1;
  vib_index[1] = vib2;
  vib_index[2] = vib3;
  vib_index[3] = vib4;
  vib_index[4] = vib5;
  vib_index[5] = vib6;
  vib_index[6] = vib7;

  vibLed = 0x00;

  range1 = (p5ft * 4) << 2;
  range2 = (p5ft * 6) << 2;
  range3 = (p5ft * 8) << 2;
  range4 = (p5ft * 10) << 2;

    // Demo
  demoControl = demoC;

  if(demoControl == 1)
  {
  	time_demo = tdemo;
	demoNum = demoN;
  }
  else
  {
  	time_demo = 0 ;
	demoNum = 0;
  }
  
//******************************************

//Start the ISRs
  sei();

}


